import datetime
from functools import wraps


class Bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def operation(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        self.operations.append(method.__name__)
        return method(self, *args, **kwargs)

    return wrapper


class QBResponse(dict):

    def __init__(self, response_type, **kwargs):
        self.response_type = response_type
        self.operations = []
        # potential to load sample data for testing
        if kwargs.get('sample_data'):
            self.update(kwargs.get('sample_data'))
        super().__init__()

    def info(self, prt=True):
        """
        Prints information about the response.
        :prt: Set to False to only grab the return as a string.
        """
        if self.response_type == 'records':
            info = []
            info.append(f'{Bcolors.OKBLUE}Sample Data:\n')
            try:
                info.append('\t' + str(self.get('data')[0]) + '\n\n')
            except KeyError as e:
                info.append('\t' + str(self.get('data')) + '\n')
            info.append(Bcolors.ENDC)
            info.append(f'{Bcolors.OKGREEN}Fields:\n')

            fields_info = []
            for field in self.get('fields'):
                field_info = []
                # build field info
                for k, v in field.items():
                    field_info.append(f'{k}: {v}')
                fields_info.append(field_info)

            # print field info
            for fi in fields_info:
                info.append('\t' + '{:<16s} {:<32s} {:<16s}\n'.format(fi[0], fi[1], fi[2]))
            info.append(Bcolors.ENDC)

            info.append(f'\n{Bcolors.OKCYAN}Metadata:\n')
            for k, v in self.get('metadata').items():
                info.append('\t{:<16s} {:<16s}\n'.format(k, str(v)))
            info.append(Bcolors.ENDC)

            info_str = ''.join(info)

            if prt:
                print(info_str)

            return info_str

    def fields(self, prop):
        """
        Gets fields, given property, i.e. label, id or type
        :param prop: property of field
        :return: list of fields
        """
        return [i.get(prop) for i in self.get('fields')]

    def data(self):
        """
        Gets data, shorthand for .get('data')
        :return: data
        """
        return self.get('data')

    def prd(self, data):
        """
        Print relevant data, simple print function.
        :param data: data to print
        """
        print("\033[91m", 'Relevant Data:\n\n', data, '\n', "\033[0m")

    @operation
    def denest(self):
        """
        Denests data, i.e. if in {'key': {'value': actual_value}} format. -> {'key': actual_value}
        :return: QBResponse
        """
        data = self.get('data')
        # ugly handling for now, need to fix
        if type(data) is dict:
            new_records = {}
            for record, record_value in data.items():
                new_record_value = {}
                for k, v in record_value.items():
                    new_record = {}
                    new_record.update({k: v.get('value')})
                    new_record_value.update(new_record)
                new_records.update({record: new_record_value})

            self.update({'data': new_records})

        if type(data) is list:
            for d in data:
                for k, v in d.items():
                    d.update({k: v.get('value')})

        return self

    @operation
    def orient(self, orient, **kwargs):
        """
        Orients the data, given orientation argument.
        :param orient: type of orientation, 'records' is the only orientation for now.
        :param kwargs: 'records' argument needs 'key' argument, to determine key for records.
        :return: QBResponse
        """
        if orient == 'records':
            if not kwargs.get('key'):
                raise ValueError('Missing required "key" argument for records orientation')
            else:
                if isinstance(kwargs.get('key'), int):
                    selector = kwargs.get('key')
                    selector = str(selector)
                else:
                    raise TypeError(f'Key must be an {int}')

            # loop data list, create dict
            records = {}
            try:
                for i in self.get('data'):
                    if 'denest' in self.operations:
                        key = i.pop(selector)
                    else:
                        key = i.pop(selector).get('value')
                    records.update({key: i})
            except KeyError as e:
                print('KeyError:', e, 'Attempting to use Record ID#')
                self.prd(self.get('data')[0])

                # attempt to find selector in fields
                for i in self.get('fields'):
                    if i.get('id') == int(selector):
                        selector = i.get('label')
                        break

                for i in self.get('data'):
                    key = i.pop(selector)
                    records.update({key: i})

            # set data to records
            self.update({'data': records})

            return self

        else:
            raise ValueError(f'{orient} is not a valid orientation.')

    def currency(self, currency_type):
        return self

    def transform(self, transformation, **kwargs):
        """
        Transforms the data, given a transformation argument.
        :param transformation: type of transformation.
        :return: QBResponse
        """
        if transformation == 'labels':

            # transform fids into labels
            if not self.get('data'):
                raise KeyError(
                    'Try transforming the data before applying additional methods.')

            data = self.get('data')
            fields = self.get('fields')

            fields_dict = {i['id']: i for i in fields}

            # replace record id numbers with record labels
            records = []
            for idx, d in enumerate(data):
                record = {}
                for k, v in d.items():
                    record.update({
                        fields_dict[int(k)].get('label'): v if v.get('value') is None else v.get('value')
                    })

                records.append(record)
            self.update({'data': records})

        if transformation == 'intround':
            raise ValueError('This transformation has been deprecated, please use the "round_ints()" method.')

        return self

    def round_ints(self):
        """
        Round int fields (remove floating .0 from int fields)
        :return:
        """
        for f in self.get('fields'):
            fid = str(f.get('id'))
            if f.get('type') == 'numeric':

                if 'denest' in self.operations:
                    for d in self.get('data'):
                        if d.get(fid).is_integer():
                            d.update({fid: int(round(d.get(fid)))})
                else:
                    for d in self.get('data'):
                        d.update({fid: {
                            'value': int(round(d.get(fid).get('value')))}})

        return self

    def convert_type(self, field_type, **kwargs):
        """
        Converts data of certain field types to standard python objects
        :param field_type:
        :param kwargs:
        :return:
        """

        def c_datetime():
            """
            Helper function that converts date time QB fields to datetime ojects.
            """
            for f in self.get('fields'):
                fid = str(f.get('id'))
                if f.get('type') == 'date time':

                    if 'denest' in self.operations:
                        for d in self.get('data'):
                            d.update({fid: datetime.datetime.strptime(d.get(fid), '%Y-%m-%dT%H:%M:%S.%fZ')})
                    else:
                        for d in self.get('data'):
                            d.update({fid: {
                                'value': datetime.datetime.strptime(d.get(fid).get('value'), '%Y-%m-%dT%H:%M:%S.%fZ')}})

        def c_currency(currency_format):

            for f in self.get('fields'):
                fid = str(f.get('id'))
                if f.get('type') == 'numeric currency':

                    if 'denest' in self.operations:
                        for d in self.get('data'):
                            fmt = currency_format + "{:,.2f}".format(d.get(fid))
                            d.update({fid: fmt})
                    else:
                        for d in self.get('data'):
                            fmt = currency_format + "{:,.2f}".format(d.get(fid).get('value'))
                            d.update({fid: {'value':  fmt}})

        if field_type == 'datetime':
            c_datetime()
        if field_type == 'numeric currency' or field_type == 'currency':
            c_currency(kwargs.get('fmt'))

        return self



import base64


class IncorrectParameters(Exception):
    def __init__(self, value, expected):
        self.value = value
        self.message = f'{value} is type {type(value)}.  Expected {expected}.'
        super().__init__(self.message)


def build_query_str(fid, operator, value):
    return f'{{{fid}.{operator}.{value}}}'


class QuickbaseParameter:

    def build(self):
        return 'Default String'

    def __str__(self):
        return self.build()


class Where(QuickbaseParameter):
    def __init__(self, fid: any, operator: str, value: any, **kwargs):
        """
        Initializes the quickbase query.
        :param fid: Field Id, represents "where"
        :param operator: Quickbase Query Language operator. https://help.quickbase.com/api-guide/do_query.html#queryOperators
        :param value: Value to compare against
        :param kwargs:
        """
        self.fid = fid
        self.operator = operator
        self.value = value

        # add quotes for strings
        if isinstance(self.value, str):
            self.value = '"{}"'.format(self.value)

    def build(self, **kwargs):

        if kwargs.get('join'):
            # check that value is a list
            if isinstance(self.value, list) is False:
                raise IncorrectParameters(self.value, list)

            # build queries
            queries = []
            for v in self.value:
                queries.append(build_query_str(self.fid, self.operator, v))

            # return chain of queries
            return f"{kwargs.get('join')}".join(queries)

        return build_query_str(self.fid, self.operator, self.value)


class Sort(QuickbaseParameter):
    def __init__(self, sort_pairs: list):
        """
        Creates a sort parameter for quickbase query.
        :param sort_pairs: tuples of fid and order, i.e. (4, 'ASC')
        """
        self.sort_pairs = sort_pairs

    def build(self):

        # ensure pairs is a list
        if not isinstance(self.sort_pairs, list):
            raise IncorrectParameters(self.sort_pairs, list)

        sorters = []
        for pair in self.sort_pairs:
            sorters.append({'fieldId': pair[0], 'order': pair[1]})

        return sorters


class Group(QuickbaseParameter):
    def __init__(self, group_pairs: list):
        """
        Creates a groupBy parameter for quickbase query.
        :param group_pairs: list of tuples, i.e. (3, 'equal-value')
        """
        self.group_pairs = group_pairs

    def build(self):

        # ensure pairs is a list
        if not isinstance(self.group_pairs, list):
            raise IncorrectParameters(self.group_pairs, list)

        sorters = []
        for pair in self.group_pairs:
            sorters.append({'fieldId': pair[0], 'grouping': pair[1]})

        return sorters


class FileUpload(dict):
    """
    Represents a file object for easy upload to quickbase.
    When uploading, set FID to FileUpload directly, bypass {'value': etc}
    """

    def __init__(self, path: str):
        """
        :param path: path to file
        """
        super().__init__()
        self.path = path

        with open(path, 'rb') as f:
            # get file as a b64 string for upload to quickbase
            file = base64.b64encode(f.read()).decode()
            self.update({'value': {'fileName': f'{f.name.split("/")[-1]}', 'data': file}})


class File(dict):
    """
    Represents a file object for easy upload to quickbase.
    When uploading, set FID to FileUpload directly, bypass {'value': etc}
    """

    def __init__(self, path: str):
        """
        :param path: path to file
        """
        super().__init__()
        self.path = path

        with open(path, 'rb') as f:
            # get file as a b64 string for upload to quickbase
            file = base64.b64encode(f.read()).decode()
            self.update({'value': {'fileName': f'{f.name.split("/")[-1]}', 'data': file}})



from quickbase_json.client import QuickbaseJSONClient


class LinkedRecord:
    def __init__(self, table, rid, table_name='default', fid=3, **kwargs):
        self.tables = {}
        self.attributes = {}
        self.add_link(table, rid, table_name, fid)

    def add_link(self, table_id, rid, table_name='default', fid=3):
        if table_name == 'default':
            table_name = table_id
        self.tables.update({table_id: {'rid': rid, 'fid': fid, 'alias': table_name}})

    def add_attribute(self, key, value):
        self.attributes.update({key: value})

    def build(self, use_aliases=False):
        pass


class LinkedRecordFactory:
    def __init__(self, realm, auth_token, table_id):
        self.realm = realm
        self.table_id = table_id
        self.records = {}
        self.client = QuickbaseJSONClient(realm, auth_token)

    def set_root(self, table, fid, where):
        records = self.client.query_records(table, fid, where)
        self.records = records
        return records

    def add_link(self, source_table, fid, table):
        pass

    def add_links(self, links):
        pass



import requests

from quickbase_json.qb_response import QBResponse


class QuickbaseJSONClient:
    def __init__(self, realm, auth, **kwargs):
        """
        Creates a client object.
        :param realm: quickbase realm
        :param auth: quickbase user token
        :param kwargs:
        """
        self.realm = realm
        self.auth = auth
        self.headers = {
            'QB-Realm-Hostname': f'{self.realm}.quickbase.com',
            'User-Agent': '{User-Agent}',
            'Authorization': f'QB-USER-TOKEN {auth}'
        }
        self.debug = True if kwargs.get('debug') else False

    """
    Records API
    """

    def query_records(self, table: str, select: list, where: str, **kwargs):
        """
        Queries for record data.
        https://developer.quickbase.com/operation/runQuery
        :param table: quickbase table
        :param select: list, list of fids to query
        :param where: Quickbase query language string. i.e. {3.EX.100}
        :param kwargs: optional request parameters.
        :return: json data of records {data: ..., fields: ...}
        """

        # create request body
        body = {'from': table, 'select': select, 'where': where}
        # update with keyword args
        body.update(kwargs)

        # add optional args
        body.update(kwargs)
        r = requests.post('https://api.quickbase.com/v1/records/query', headers=self.headers, json=body).json()

        # create response object
        res = QBResponse('records')

        # update response object with JSON data from request
        res.update(r)
        return res

    def insert_update_records(self, table: str, data: list):
        """
        Inerts or updates records in a given table.
        https://developer.quickbase.com/operation/upsert
        :param table: table to add records to
        :param data: list of dict of data, [{"6": {"value": 'example'}}] (do not include 3, record id to insert)
        :return: record id of created/updated records
        """

        body = {'to': table, 'data': data}
        return requests.post('https://api.quickbase.com/v1/records', headers=self.headers, json=body).json()

    def delete_records(self, table: str, where: str):
        """
        Deletes records in a table based on a query.
        https://developer.quickbase.com/operation/deleteRecords
        :param table: The unique identifier of the table.
        :param where: The filter to delete records. To delete all records specify a filter that will include all records.
        :return: dict, numberDeleted is number of records deleted.
        """

        headers = self.headers
        body = {'to': table, 'where': where}
        return requests.delete('https://api.quickbase.com/v1/records', headers=headers, json=body).json()

    """
    Table API
    """

    def create_table(self, app_id: str, name: str, **kwargs):
        """
        Creates a table in an application.
        https://developer.quickbase.com/operation/createTable
        :param app_id: The unique identifier of an app
        :param name: name of table
        :param kwargs: optional args
        :return:
        """

        headers = self.headers
        params = {'appId': f'{app_id}'}
        body = {'name': name}.update(kwargs)
        return requests.post('https://api.quickbase.com/v1/tables', params=params, headers=headers, json=body).json()

    def get_tables(self, app_id: str):
        """
        Gets all tables in an application
        https://developer.quickbase.com/operation/getAppTables
        :param app_id: The unique identifier of an app.
        :return: dict of all tables in application.
        """

        headers = self.headers
        params = {'appId': f'{app_id}'}
        body = None
        return requests.post('https://api.quickbase.com/v1/tables', params=params, headers=headers, json=body).json()

    """
    Fields API
    """

    def get_fields(self, table_id: str, **kwargs):
        """
        Get fields for a given table.
        https://developer.quickbase.com/operation/getFields
        :param table_id: Id of quickbase table
        :param kwargs: optional args
        :return:
        """

        headers = self.headers
        params = {'tableId': f'{table_id}'}
        return requests.get('https://api.quickbase.com/v1/tables', params=params, headers=headers).json()

    """
    Misc.
    """
    def __str__(self):
        return f'Quickbase Client: {self.realm}'
